# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions for creating model backbones"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import copy

from tensorflow.keras import backend as K
from tensorflow.keras import applications
from tensorflow.keras.backend import is_keras_tensor
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Conv2D, Conv3D, BatchNormalization
from tensorflow.keras.layers import Activation, MaxPool2D, MaxPool3D
from tensorflow.keras.layers import TimeDistributed
from tensorflow.keras.utils import get_source_inputs


def featurenet_block(x, n_filters):
    """Add a set of layers that make up one unit of the featurenet backbone

    Args:
        x (tensorflow.keras.Layer): Keras layer object to pass to
            backbone unit
        n_filters (int): Number of filters to use for convolutional layers

    Returns:
        tensorflow.keras.Layer: Keras layer object
    """
    df = K.image_data_format()
    # conv set 1
    x = Conv2D(n_filters, (3, 3), strides=(1, 1), padding='same', data_format=df)(x)
    x = BatchNormalization(axis=-1)(x)
    x = Activation('relu')(x)
    # conv set 2
    x = Conv2D(n_filters, (3, 3), strides=(1, 1), padding='same', data_format=df)(x)
    x = BatchNormalization(axis=-1)(x)
    x = Activation('relu')(x)
    # Final max pooling stage
    x = MaxPool2D(pool_size=(2, 2), padding='same', data_format=df)(x)

    return x


def featurenet_3D_block(x, n_filters):
    """Add a set of layers that make up one unit of the featurenet 3D backbone

    Args:
        x (tensorflow.keras.Layer): Keras layer object to pass to
            backbone unit
        n_filters (int): Number of filters to use for convolutional layers

    Returns:
        tensorflow.keras.Layer: Keras layer object
    """
    df = K.image_data_format()
    # conv set 1
    x = Conv3D(n_filters, (3, 3, 3), strides=(1, 1, 1), padding='same', data_format=df)(x)
    x = BatchNormalization(axis=-1)(x)
    x = Activation('relu')(x)
    # conv set 2
    x = Conv3D(n_filters, (3, 3, 3), strides=(1, 1, 1), padding='same', data_format=df)(x)
    x = BatchNormalization(axis=-1)(x)
    x = Activation('relu')(x)
    # Final max pooling stage
    x = MaxPool3D(pool_size=(2, 2, 2), data_format=df)(x)

    return x


def featurenet_backbone(input_tensor=None, input_shape=None,
                        n_filters=32, **kwargs):
    """Construct the deepcell backbone with five convolutional units

    Args:
        input_tensor (tensor): Input tensor to specify input size
        n_filters (int): Number of filters for convolutional layers

    Returns:
        tuple: List of backbone layers, list of backbone names
    """
    if input_tensor is None:
        img_input = Input(shape=input_shape)
    else:
        if not is_keras_tensor(input_tensor):
            img_input = Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    # Build out backbone
    c1 = featurenet_block(img_input, n_filters)  # 1/2 64x64
    c2 = featurenet_block(c1, n_filters)  # 1/4 32x32
    c3 = featurenet_block(c2, n_filters)  # 1/8 16x16
    c4 = featurenet_block(c3, n_filters)  # 1/16 8x8
    c5 = featurenet_block(c4, n_filters)  # 1/32 4x4

    backbone_features = [c1, c2, c3, c4, c5]
    backbone_names = ['C1', 'C2', 'C3', 'C4', 'C5']
    output_dict = {}
    for name, feature in zip(backbone_names, backbone_features):
        output_dict[name] = feature

    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)
    else:
        inputs = img_input

    model = Model(inputs=inputs, outputs=backbone_features)
    return model, output_dict


def featurenet_3D_backbone(input_tensor=None, input_shape=None,
                           n_filters=32, **kwargs):
    """Construct the deepcell backbone with five convolutional units

    Args:
        input_tensor (tensor): Input tensor to specify input size
        n_filters (int): Number of filters for convolutional layers

    Returns:
        tuple: List of backbone layers, list of backbone names
    """
    if input_tensor is None:
        img_input = Input(shape=input_shape)
    else:
        if not is_keras_tensor(input_tensor):
            img_input = Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    # Build out backbone
    c1 = featurenet_3D_block(img_input, n_filters)  # 1/2 64x64
    c2 = featurenet_3D_block(c1, n_filters)  # 1/4 32x32
    c3 = featurenet_3D_block(c2, n_filters)  # 1/8 16x16
    c4 = featurenet_3D_block(c3, n_filters)  # 1/16 8x8
    c5 = featurenet_3D_block(c4, n_filters)  # 1/32 4x4

    backbone_features = [c1, c2, c3, c4, c5]
    backbone_names = ['C1', 'C2', 'C3', 'C4', 'C5']
    output_dict = {}
    for name, feature in zip(backbone_names, backbone_features):
        output_dict[name] = feature

    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)
    else:
        inputs = img_input

    model = Model(inputs=inputs, outputs=backbone_features)
    return model, output_dict


def get_backbone(backbone, input_tensor=None, input_shape=None,
                 use_imagenet=False, return_dict=True,
                 frames_per_batch=1, **kwargs):
    """Retrieve backbones for the construction of feature pyramid networks.

    Args:
        backbone (str): Name of the backbone to be retrieved.
        input_tensor (tensor): The input tensor for the backbone.
            Should have channel dimension of size 3
        use_imagenet (bool): Load pre-trained weights for the backbone
        return_dict (bool): Whether to return a dictionary of backbone layers,
            e.g. ``{'C1': C1, 'C2': C2, 'C3': C3, 'C4': C4, 'C5': C5}``.
            If false, the whole model is returned instead
        kwargs (dict): Keyword dictionary for backbone constructions.
            Relevant keys include ``'include_top'``,
            ``'weights'`` (should be ``None``),
            ``'input_shape'``, and ``'pooling'``.

    Returns:
        tensorflow.keras.Model: An instantiated backbone

    Raises:
        ValueError: bad backbone name
        ValueError: featurenet backbone with pre-trained imagenet
    """
    _backbone = str(backbone).lower()

    featurenet_backbones = {
        'featurenet': featurenet_backbone,
        'featurenet3d': featurenet_3D_backbone,
        'featurenet_3d': featurenet_3D_backbone
    }
    vgg_backbones = {
        'vgg16': applications.vgg16.VGG16,
        'vgg19': applications.vgg19.VGG19,
    }
    densenet_backbones = {
        'densenet121': applications.densenet.DenseNet121,
        'densenet169': applications.densenet.DenseNet169,
        'densenet201': applications.densenet.DenseNet201,
    }
    mobilenet_backbones = {
        'mobilenet': applications.mobilenet.MobileNet,
        'mobilenetv2': applications.mobilenet_v2.MobileNetV2,
        'mobilenet_v2': applications.mobilenet_v2.MobileNetV2
    }
    resnet_backbones = {
        'resnet50': applications.resnet.ResNet50,
        'resnet101': applications.resnet.ResNet101,
        'resnet152': applications.resnet.ResNet152,
    }
    resnet_v2_backbones = {
        'resnet50v2': applications.resnet_v2.ResNet50V2,
        'resnet101v2': applications.resnet_v2.ResNet101V2,
        'resnet152v2': applications.resnet_v2.ResNet152V2,
    }
    # resnext_backbones = {
    #     'resnext50': applications.resnext.ResNeXt50,
    #     'resnext101': applications.resnext.ResNeXt101,
    # }
    nasnet_backbones = {
        'nasnet_large': applications.nasnet.NASNetLarge,
        'nasnet_mobile': applications.nasnet.NASNetMobile,
    }
    efficientnet_backbones = {
        'efficientnetb0': applications.efficientnet.EfficientNetB0,
        'efficientnetb1': applications.efficientnet.EfficientNetB1,
        'efficientnetb2': applications.efficientnet.EfficientNetB2,
        'efficientnetb3': applications.efficientnet.EfficientNetB3,
        'efficientnetb4': applications.efficientnet.EfficientNetB4,
        'efficientnetb5': applications.efficientnet.EfficientNetB5,
        'efficientnetb6': applications.efficientnet.EfficientNetB6,
        'efficientnetb7': applications.efficientnet.EfficientNetB7,
    }

    # TODO: Check and make sure **kwargs is in the right format.
    # 'weights' flag should be None, and 'input_shape' must have size 3 on the channel axis
    if frames_per_batch == 1:
        if input_tensor is not None:
            img_input = input_tensor
        else:
            if input_shape:
                img_input = Input(shape=input_shape)
            else:
                img_input = Input(shape=(None, None, 3))
    else:
        # using 3D data but a 2D backbone.
        # TODO: why ignore input_tensor
        if input_shape:
            img_input = Input(shape=input_shape)
        else:
            img_input = Input(shape=(None, None, 3))

    if use_imagenet:
        kwargs_with_weights = copy.copy(kwargs)
        kwargs_with_weights['weights'] = 'imagenet'
    else:
        kwargs['weights'] = None

    if _backbone in featurenet_backbones:
        if use_imagenet:
            raise ValueError('A featurenet backbone that is pre-trained on '
                             'imagenet does not exist')

        model_cls = featurenet_backbones[_backbone]
        model, output_dict = model_cls(input_tensor=img_input, **kwargs)

        layer_outputs = [output_dict['C1'], output_dict['C2'], output_dict['C3'],
                         output_dict['C4'], output_dict['C5']]

    elif _backbone in vgg_backbones:
        model_cls = vgg_backbones[_backbone]
        model = model_cls(input_tensor=img_input, **kwargs)

        # Set the weights of the model if requested
        if use_imagenet:
            model_with_weights = model_cls(**kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        layer_names = ['block1_pool', 'block2_pool', 'block3_pool', 'block4_pool', 'block5_pool']
        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    elif _backbone in densenet_backbones:
        model_cls = densenet_backbones[_backbone]
        model = model_cls(input_tensor=img_input, **kwargs)
        if _backbone == 'densenet121':
            blocks = [6, 12, 24, 16]
        elif _backbone == 'densenet169':
            blocks = [6, 12, 32, 32]
        elif _backbone == 'densenet201':
            blocks = [6, 12, 48, 32]

        # Set the weights of the model if requested
        if use_imagenet:
            model_with_weights = model_cls(**kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        layer_names = ['conv1/relu'] + ['conv{}_block{}_concat'.format(idx + 2, block_num)
                                        for idx, block_num in enumerate(blocks)]
        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    elif _backbone in resnet_backbones:
        model_cls = resnet_backbones[_backbone]
        model = model_cls(input_tensor=img_input, **kwargs)

        # Set the weights of the model if requested
        if use_imagenet:
            model_with_weights = model_cls(**kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        if _backbone == 'resnet50':
            layer_names = ['conv1_relu', 'conv2_block3_out', 'conv3_block4_out',
                           'conv4_block6_out', 'conv5_block3_out']
        elif _backbone == 'resnet101':
            layer_names = ['conv1_relu', 'conv2_block3_out', 'conv3_block4_out',
                           'conv4_block23_out', 'conv5_block3_out']
        elif _backbone == 'resnet152':
            layer_names = ['conv1_relu', 'conv2_block3_out', 'conv3_block8_out',
                           'conv4_block36_out', 'conv5_block3_out']

        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    elif _backbone in resnet_v2_backbones:
        model_cls = resnet_v2_backbones[_backbone]
        model = model_cls(input_tensor=img_input, **kwargs)

        # Set the weights of the model if requested
        if use_imagenet:
            model_with_weights = model_cls(**kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        if _backbone == 'resnet50v2':
            layer_names = ['post_relu', 'conv2_block3_out', 'conv3_block4_out',
                           'conv4_block6_out', 'conv5_block3_out']
        elif _backbone == 'resnet101v2':
            layer_names = ['post_relu', 'conv2_block3_out', 'conv3_block4_out',
                           'conv4_block23_out', 'conv5_block3_out']
        elif _backbone == 'resnet152v2':
            layer_names = ['post_relu', 'conv2_block3_out', 'conv3_block8_out',
                           'conv4_block36_out', 'conv5_block3_out']

        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    # elif _backbone in resnext_backbones:
    #     model_cls = resnext_backbones[_backbone]
    #     model = model_cls(input_tensor=img_input, **kwargs)
    #
    #     # Set the weights of the model if requested
    #     if use_imagenet:
    #         model_with_weights = model_cls(**kwargs_with_weights)
    #         model_with_weights.save_weights('model_weights.h5')
    #         model.load_weights('model_weights.h5', by_name=True)
    #
    #     if _backbone == 'resnext50':
    #         layer_names = ['conv1_relu', 'conv2_block3_out', 'conv3_block4_out',
    #                        'conv4_block6_out', 'conv5_block3_out']
    #     elif _backbone == 'resnext101':
    #         layer_names = ['conv1_relu', 'conv2_block3_out', 'conv3_block4_out',
    #                        'conv4_block23_out', 'conv5_block3_out']
    #
    #     layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    elif _backbone in mobilenet_backbones:
        model_cls = mobilenet_backbones[_backbone]
        alpha = kwargs.pop('alpha', 1.0)
        model = model_cls(alpha=alpha, input_tensor=img_input, **kwargs)
        if _backbone.endswith('v2'):
            block_ids = (2, 5, 12)
            layer_names = ['expanded_conv_project_BN'] + \
                          ['block_%s_add' % i for i in block_ids] + \
                          ['block_16_project_BN']
        else:
            block_ids = (1, 3, 5, 11, 13)
            layer_names = ['conv_pw_%s_relu' % i for i in block_ids]

        # Set the weights of the model if requested
        if use_imagenet:
            model_with_weights = model_cls(alpha=alpha, **kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    elif _backbone in nasnet_backbones:
        model_cls = nasnet_backbones[_backbone]
        model = model_cls(input_tensor=img_input, **kwargs)
        if _backbone.endswith('large'):
            block_ids = [5, 12, 18]
        else:
            block_ids = [3, 8, 12]

        # Set the weights of the model if requested
        if use_imagenet:
            model_with_weights = model_cls(**kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        layer_names = ['stem_bn1', 'reduction_concat_stem_1']
        layer_names.extend(['normal_concat_%s' % i for i in block_ids])
        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    elif _backbone in efficientnet_backbones:
        model_cls = efficientnet_backbones[_backbone]
        model = model_cls(input_tensor=img_input, **kwargs)

        if use_imagenet:
            model_with_weights = model_cls(**kwargs_with_weights)
            model_with_weights.save_weights('model_weights.h5')
            model.load_weights('model_weights.h5', by_name=True)

        layer_names = ['block2a_expand_activation', 'block3a_expand_activation',
                       'block4a_expand_activation', 'block6a_expand_activation',
                       'top_activation']
        layer_outputs = [model.get_layer(name=ln).output for ln in layer_names]

    else:
        join = lambda x: [v for y in x for v in list(y.keys())]
        backbones = join([featurenet_backbones, densenet_backbones,
                          resnet_backbones, resnet_v2_backbones,
                          vgg_backbones, nasnet_backbones,
                          mobilenet_backbones, efficientnet_backbones])
        raise ValueError('Invalid value for `backbone`. Must be one of: %s' %
                         ', '.join(backbones))

    if frames_per_batch > 1:

        time_distributed_outputs = []
        for i, out in enumerate(layer_outputs):
            td_name = 'td_{}'.format(i)
            model_name = 'model_{}'.format(i)
            time_distributed_outputs.append(
                TimeDistributed(Model(model.input, out, name=model_name),
                                name=td_name)(input_tensor))

        if time_distributed_outputs:
            layer_outputs = time_distributed_outputs

    output_dict = {'C{}'.format(i + 1): j for i, j in enumerate(layer_outputs)}
    return (model, output_dict) if return_dict else model



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for Mesmer Application"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest

import numpy as np

from unittest.mock import Mock

from tensorflow.python.platform import test

from deepcell.model_zoo import PanopticNet
from deepcell.applications import Mesmer
from deepcell.applications import MultiplexSegmentation
from deepcell.applications.mesmer import format_output_mesmer
from deepcell.applications.mesmer import mesmer_postprocess
from deepcell.applications.mesmer import mesmer_preprocess


# test pre- and post-processing functions
def test_mesmer_preprocess():
    height, width = 300, 300
    img = np.random.randint(0, 100, (height, width))

    # make rank 4 (batch, X, y, channel)
    img = np.expand_dims(img, axis=0)
    img = np.expand_dims(img, axis=-1)

    # single bright spot
    img[0, 200, 200, 0] = 5000

    # histogram normalized
    processed = mesmer_preprocess(img)
    assert (processed <= 1).all() and (processed >= -1).all()

    # maxima is no longer significantly greater than rest of image
    new_spot_val = processed[0, 200, 200, 0]
    processed[0, 200, 200, 0] = 0.5
    next_max_val = np.max(processed)

    # difference between bright spot and next greatest value is essentially nothing
    assert np.round(new_spot_val / next_max_val, 1) == 1

    # histogram normalization without thresholding
    processed_hist = mesmer_preprocess(img, threshold=False)
    assert (processed_hist <= 1).all() and (processed_hist >= -1).all()

    new_spot_val = processed_hist[0, 200, 200, 0]
    processed_hist[0, 200, 200, 0] = 0.5
    next_max_val = np.max(processed_hist)
    assert np.round(new_spot_val / next_max_val, 1) > 1

    # thresholding without histogram normalization
    processed_thresh = mesmer_preprocess(img, normalize=False)
    assert not (processed_thresh <= 1).all()

    new_spot_val = processed_thresh[0, 200, 200, 0]
    processed_thresh[0, 200, 200, 0] = 0.5
    next_max_val = np.max(processed_thresh)
    assert np.round(new_spot_val / next_max_val, 1) == 1

    # no change to image
    not_processed = mesmer_preprocess(img, normalize=False, threshold=False)
    assert np.all(not_processed == img)

    # bad input
    with pytest.raises(ValueError):
        _ = mesmer_preprocess(np.zeros((3, 50, 50)))


def test_mesmer_postprocess(mocker):
    # create dict, with each image having a different constant value
    base_array = np.ones((1, 20, 20, 1))

    whole_cell_list = [base_array * mult for mult in range(1, 3)]

    nuclear_list = [base_array * mult for mult in range(3, 5)]

    model_output = {'whole-cell': whole_cell_list, 'nuclear': nuclear_list}

    # whole cell predictions only
    whole_cell = mesmer_postprocess(model_output=model_output,
                                    compartment='whole-cell')
    assert whole_cell.shape == (1, 20, 20, 1)

    # nuclear predictions only
    nuclear = mesmer_postprocess(model_output=model_output,
                                 compartment='nuclear')
    assert nuclear.shape == (1, 20, 20, 1)

    # both whole-cell and nuclear predictions
    both = mesmer_postprocess(model_output=model_output,
                              compartment='both')
    assert both.shape == (1, 20, 20, 2)

    # make sure correct arrays are being passed to helper function
    def mock_deep_watershed(model_output):
        pixelwise_interior_vals = model_output[-1]
        return pixelwise_interior_vals

    mocker.patch('deepcell.applications.mesmer.deep_watershed',
                 mock_deep_watershed)

    # whole cell predictions only
    whole_cell_mocked = mesmer_postprocess(model_output=model_output,
                                           compartment='whole-cell')

    assert np.array_equal(whole_cell_mocked, whole_cell_list[1])

    # nuclear predictions only
    whole_cell_mocked = mesmer_postprocess(model_output=model_output,
                                           compartment='nuclear')

    assert np.array_equal(whole_cell_mocked, nuclear_list[1])

    with pytest.raises(ValueError):
        whole_cell = mesmer_postprocess(model_output=model_output,
                                        compartment='invalid')


def test_format_output_mesmer():

    # create output list, each with a different constant value across image
    base_array = np.ones((1, 20, 20, 1))

    whole_cell_list = [base_array * mult for mult in range(1, 5)]
    whole_cell_list = [whole_cell_list[0],
                       np.concatenate(whole_cell_list[1:4], axis=-1)]

    # create output list for nuclear predictions
    nuclear_list = [img * 2 for img in whole_cell_list]

    combined_list = whole_cell_list + nuclear_list

    output = format_output_mesmer(combined_list)

    assert set(output.keys()) == {'whole-cell', 'nuclear'}

    assert np.array_equal(output['whole-cell'][0], base_array)
    assert np.array_equal(output['nuclear'][0], base_array * 2)

    assert np.array_equal(output['whole-cell'][1], base_array * 3)
    assert np.array_equal(output['nuclear'][1], base_array * 6)

    with pytest.raises(ValueError):
        output = format_output_mesmer(combined_list[:3])


# test application
class TestMesmer(test.TestCase):

    def test_mesmer_app(self):
        with self.cached_session():
            whole_cell_classes = [1, 3]
            nuclear_classes = [1, 3]
            num_semantic_classes = whole_cell_classes + nuclear_classes
            num_semantic_heads = len(num_semantic_classes)

            model = PanopticNet(
                'resnet50',
                input_shape=(256, 256, 2),
                norm_method=None,
                num_semantic_heads=num_semantic_heads,
                num_semantic_classes=num_semantic_classes,
                location=True,
                include_top=True,
                use_imagenet=False)

            app = Mesmer(model)

            # test output shape
            shape = app.model.output_shape
            self.assertIsInstance(shape, list)
            self.assertEqual(len(shape), 4)

            # test predict with default
            x = np.random.rand(1, 500, 500, 2)
            y = app.predict(x)
            self.assertEqual(x.shape[:-1], y.shape[:-1])

            # test predict with nuclear compartment only
            x = np.random.rand(1, 500, 500, 2)
            y = app.predict(x, compartment='nuclear')
            self.assertEqual(x.shape[:-1], y.shape[:-1])
            self.assertEqual(y.shape[-1], 1)

            # test predict with cell compartment only
            x = np.random.rand(1, 500, 500, 2)
            y = app.predict(x, compartment='whole-cell')
            self.assertEqual(x.shape[:-1], y.shape[:-1])
            self.assertEqual(y.shape[-1], 1)

            # test predict with both cell and nuclear compartments
            x = np.random.rand(1, 500, 500, 2)
            y = app.predict(x, compartment='both')
            self.assertEqual(x.shape[:-1], y.shape[:-1])
            self.assertEqual(y.shape[-1], 2)

            # test that kwargs are passed through successfully
            app._predict_segmentation = Mock()

            # get defaults
            _ = app.predict(x, compartment='whole-cell')
            args = app._predict_segmentation.call_args[1]
            default_cell_kwargs = args['postprocess_kwargs']['whole_cell_kwargs']
            default_nuc_kwargs = args['postprocess_kwargs']['nuclear_kwargs']

            # change one of the args for each compartment
            maxima_threshold_cell = default_cell_kwargs['maxima_threshold'] + 0.1
            radius_nuc = default_nuc_kwargs['radius'] + 2

            _ = app.predict(x, compartment='whole-cell',
                            postprocess_kwargs_whole_cell={'maxima_threshold':
                                                           maxima_threshold_cell},
                            postprocess_kwargs_nuclear={'radius': radius_nuc})

            args = app._predict_segmentation.call_args[1]
            cell_kwargs = args['postprocess_kwargs']['whole_cell_kwargs']
            assert cell_kwargs['maxima_threshold'] == maxima_threshold_cell

            nuc_kwargs = args['postprocess_kwargs']['nuclear_kwargs']
            assert nuc_kwargs['radius'] == radius_nuc

            # check that rest are unchanged
            cell_kwargs['maxima_threshold'] = default_cell_kwargs['maxima_threshold']
            assert cell_kwargs == default_cell_kwargs

            nuc_kwargs['radius'] = default_nuc_kwargs['radius']
            assert nuc_kwargs == default_nuc_kwargs

            # test legacy version
            old_app = MultiplexSegmentation(model)

            # test predict with default
            x = np.random.rand(1, 500, 500, 2)
            y = old_app.predict(x)
            self.assertEqual(x.shape[:-1], y.shape[:-1])



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for the tensor product layers"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
import tensorflow as tf

from tensorflow.python import keras
from tensorflow.python.keras import keras_parameterized
from tensorflow.python.keras import testing_utils
from tensorflow.python.platform import test

from deepcell import layers


@keras_parameterized.run_all_keras_modes
class TensorProdTest(keras_parameterized.TestCase):

    def test_tensorproduct(self):
        custom_objects = {'TensorProduct': layers.TensorProduct}
        with tf.keras.utils.custom_object_scope(custom_objects):
            testing_utils.layer_test(
                layers.TensorProduct,
                kwargs={'output_dim': 3},
                input_shape=(3, 2))

            testing_utils.layer_test(
                layers.TensorProduct,
                kwargs={'output_dim': 3},
                input_shape=(3, 4, 2))

            testing_utils.layer_test(
                layers.TensorProduct,
                kwargs={'output_dim': 3},
                input_shape=(None, None, 2))

            testing_utils.layer_test(
                layers.TensorProduct,
                kwargs={'output_dim': 3},
                input_shape=(3, 4, 5, 2))

            testing_utils.layer_test(
                layers.TensorProduct,
                kwargs={'output_dim': 3,
                        'data_format': 'channels_first'},
                input_shape=(3, 2, 4, 5))

            # test no bias
            testing_utils.layer_test(
                layers.TensorProduct,
                kwargs={'output_dim': 2,
                        'use_bias': False},
                input_shape=(3, 5, 6, 4))

            # test bad input channel
            with self.assertRaises(ValueError):
                testing_utils.layer_test(
                    layers.TensorProduct,
                    kwargs={'output_dim': 3},
                    input_shape=(3, 5, 6, None))

    def test_tensorproduct_regularization(self):
        layer = layers.TensorProduct(
            3,
            kernel_regularizer=tf.keras.regularizers.l1(0.01),
            bias_regularizer='l1',
            activity_regularizer='l2',
            name='tensorproduct_reg')
        layer(tf.keras.backend.variable(np.ones((2, 4))))
        self.assertEqual(3, len(layer.losses))

    def test_tensorproduct_constraints(self):
        k_constraint = tf.keras.constraints.max_norm(0.01)
        b_constraint = tf.keras.constraints.max_norm(0.01)
        layer = layers.TensorProduct(
            3,
            kernel_constraint=k_constraint,
            bias_constraint=b_constraint)
        layer(tf.keras.backend.variable(np.ones((2, 4))))
        self.assertEqual(layer.kernel.constraint, k_constraint)
        self.assertEqual(layer.bias.constraint, b_constraint)


if __name__ == '__main__':
    test.main()



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Dataset Builders"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import numpy as np
import tensorflow as tf
import tensorflow_addons as tfa

from deepcell.data import split_dataset


def temporal_slice(X, y, track_length=8):
    """Randomly slice movies and labels with a length of ``track_length``.

    Args:
        X (dict): Dictionary of feature data.
        y (dict): Dictionary of labels.
        track_length (int): Length of temporal slices.

    Returns:
        tuple(dict, dict): Tuple of sliced ``X`` and ``y`` data.
    """
    appearances = X['appearances']
    max_time = tf.shape(appearances)[0] - track_length

    t_start = tf.random.uniform(shape=[], minval=0,
                                maxval=max_time,
                                dtype=tf.int32)

    t_end = t_start + track_length

    for key, data in X.items():
        X[key] = data[t_start:t_end]

    for key, data in y.items():
        y[key] = data[t_start:t_end - 1]

    return (X, y)


def random_rotate(X, y, rotation_range=0):
    """Randomly rotate centroids and appearances.

    Args:
        X (dict): Dictionary of feature data.
        rotation_range (int): Maximum rotation range in degrees.

    Returns:
        dict: Rotated ``X`` data.
    """
    appearances = X['appearances']
    centroids = X['centroids']

    # Calculate the random rotation in radians
    rg = rotation_range * math.pi / 180
    theta = tf.random.uniform(shape=[1], minval=-rg, maxval=rg)

    # Transform appearances
    old_shape = tf.shape(appearances)
    new_shape = [-1, old_shape[2], old_shape[3], old_shape[4]]
    img = tf.reshape(appearances, new_shape)
    img = tfa.image.rotate(img, theta)
    img = tf.reshape(img, old_shape)
    X['appearances'] = img

    # Rotate coordinates
    cos_theta = tf.math.cos(theta)
    sin_theta = tf.math.sin(theta)
    rotation_matrix = tf.concat([
        [cos_theta, -sin_theta],
        [sin_theta, cos_theta],
    ], axis=1)
    transformed_centroids = tf.matmul(centroids, rotation_matrix)
    X['centroids'] = transformed_centroids

    return X, y


def random_translate(X, y, range=512):
    """Randomly translate the centroids."""
    centroids = X['centroids']
    r0 = tf.random.uniform([1, 1, 2], -range, range)
    transformed_centroids = centroids + r0
    X['centroids'] = transformed_centroids
    return X, y


def prepare_dataset(track_info, batch_size=32, buffer_size=256,
                    seed=None, track_length=8, rotation_range=0,
                    translation_range=0, val_size=0.2, test_size=0):
    """Build and prepare the tracking dataset.

    Args:
        track_info (dict): A dictionary of all input and output features
        batch_size (int): number of examples per batch
        buffer_size (int): number of samples to buffer
        seed (int): Random seed
        track_length (int): Number of frames per example
        rotation_range (int): Maximum degrees to rotate inputs
        translation_range (int): Maximum range of translation,
            should be equivalent to original input image size.
        val_size (float): Fraction of data to split into validation set.
        test_size (float): Fraction of data to split into test set.

    Returns:
        tf.data.Dataset: A ``tf.data.Dataset`` object ready for training.
    """
    input_dict = {
        'appearances': track_info['appearances'],
        'centroids': track_info['centroids'],
        'morphologies': track_info['morphologies'],
        'adj_matrices': track_info['norm_adj_matrices'],
    }

    output_dict = {'temporal_adj_matrices': track_info['temporal_adj_matrices']}

    dataset = tf.data.Dataset.from_tensor_slices((input_dict, output_dict))

    dataset = dataset.shuffle(buffer_size, seed=seed).repeat()

    # randomly sample along the temporal axis
    sample = lambda X, y: temporal_slice(X, y, track_length=track_length)
    dataset = dataset.map(sample, num_parallel_calls=tf.data.AUTOTUNE)

    # split into train/val before doing any augmentation
    train_data, val_data, test_data = split_dataset(dataset, val_size, test_size)

    # randomly rotate
    rotate = lambda X, y: random_rotate(X, y, rotation_range=rotation_range)
    train_data = train_data.map(rotate, num_parallel_calls=tf.data.AUTOTUNE)

    # randomly translate centroids
    translate = lambda X, y: random_translate(X, y, range=translation_range)
    train_data = train_data.map(translate, num_parallel_calls=tf.data.AUTOTUNE)

    # batch the data
    train_data = train_data.batch(batch_size)
    val_data = val_data.batch(batch_size)
    test_data = test_data.batch(batch_size)

    # prefetch the data
    train_data = train_data.prefetch(tf.data.AUTOTUNE)
    val_data = val_data.prefetch(tf.data.AUTOTUNE)
    test_data = test_data.prefetch(tf.data.AUTOTUNE)

    return train_data, val_data, test_data



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for the upsampling layers"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
from tensorflow.keras import backend as K
from tensorflow.python.keras import keras_parameterized
from tensorflow.python.keras import testing_utils
from tensorflow.keras.utils import custom_object_scope
from tensorflow.python.platform import test

from deepcell import layers


@keras_parameterized.run_all_keras_modes
class TestComparison(keras_parameterized.TestCase):
    def test_simple(self):
        # create simple Comparison layer
        comparison_layer = layers.Comparison()

        # create input before
        before_objs = 7
        before_chan = 2
        before = np.zeros((1, 5, before_objs, before_chan), dtype=K.floatx())
        before = K.variable(before)
        after_objs = 8
        after_chan = 3
        after = np.zeros((1, 5, after_objs, after_chan), dtype=K.floatx())
        after = K.variable(after)

        target_chan = before_chan + after_chan
        target_shape = tuple(list(before.shape[:-1]) + [after_objs, target_chan])
        target = np.zeros(target_shape, dtype=K.floatx())
        expected = target
        target = K.variable(target)

        # compute output
        computed_shape = comparison_layer.compute_output_shape(
            [before.shape, after.shape])

        actual = comparison_layer.call([before, after])
        actual = K.get_value(actual)

        self.assertEqual(actual.shape, computed_shape)
        self.assertAllEqual(actual, expected)


@keras_parameterized.run_all_keras_modes
class TestDeltaReshape(keras_parameterized.TestCase):
    def test_simple(self):
        # create simple DeltaReshape layer
        delta_reshape_layer = layers.DeltaReshape()

        # create input before
        before_objs = 7
        before = np.zeros((3, 5, before_objs, 1), dtype=K.floatx())
        before = K.variable(before)
        after_objs = 8
        after = np.zeros((1, 1, after_objs, 1), dtype=K.floatx())
        after = K.variable(after)

        target_shape = tuple(list(before.shape[:-1]) + list(after.shape[-2:]))
        target = np.zeros(target_shape, dtype=K.floatx())
        expected = target
        target = K.variable(target)

        # compute output
        computed_shape = delta_reshape_layer.compute_output_shape(
            [before.shape, after.shape])

        actual = delta_reshape_layer.call([before, after])
        actual = K.get_value(actual)

        self.assertEqual(actual.shape, computed_shape)
        self.assertAllEqual(actual, expected)


@keras_parameterized.run_all_keras_modes
class TestUnmerge(keras_parameterized.TestCase):

    def test_unmerge(self):
        track_length = 5
        max_cells = 10
        embedding_dim = 64

        custom_objects = {'Unmerge': layers.Unmerge}
        with self.cached_session():
            with custom_object_scope(custom_objects):
                testing_utils.layer_test(
                    layers.Unmerge,
                    kwargs={'track_length': track_length,
                            'max_cells': max_cells,
                            'embedding_dim': embedding_dim},
                    input_shape=(None, track_length * max_cells, embedding_dim))


@keras_parameterized.run_all_keras_modes
class TestTemporalMerge(keras_parameterized.TestCase):

    def test_temporal_merge(self):
        track_length = 5
        max_cells = 7
        encoder_dim = 32

        custom_objects = {'TemporalMerge': layers.TemporalMerge}
        with self.cached_session():
            with custom_object_scope(custom_objects):
                testing_utils.layer_test(
                    layers.TemporalMerge,
                    kwargs={'encoder_dim': encoder_dim},
                    input_shape=(None, track_length, max_cells, encoder_dim))



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

from codecs import open

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(here, 'README.md'), 'r', 'utf-8') as f:
    readme = f.read()


about = {}
with open(os.path.join(here, 'deepcell', '_version.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)


setup(
    name=about['__title__'],
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    description=about['__description__'],
    url=about['__url__'],
    download_url=about['__download_url__'],
    license=about['__license__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy>=1.16.6,<1.20.0',
        'pydot>=1.4.2,<2',
        'scipy>=1.2.3,<2',
        'scikit-image>=0.14.5',
        'scikit-learn>=0.20.4',
        'tensorflow-cpu~=2.5.1',
        'tensorflow_addons~=0.13.0',
        'spektral~=1.0.4',
        'jupyter>=1.0.0,<2',
        'opencv-python-headless<5',
        'deepcell-tracking~=0.4.4',
        'deepcell-toolbox~=0.10.0'
    ],
    extras_require={
        'tests': [
            'pytest<6',
            'pytest-cov',
            'pytest-pep8',
        ],
    },
    packages=find_packages(),
    python_requires='>=3.6, <3.10',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Miscellaneous utility functions"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import re


def sorted_nicely(l):
    """Sort a list of strings by the numerical order of all substrings

    Args:
        l (list): List of strings to sort

    Returns:
        list: a sorted list
    """
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)


def get_sorted_keys(dict_to_sort):
    """Gets the keys from a dict and sorts them in ascending order.
    Assumes keys are of the form ``Ni``, where ``N`` is a letter and ``i``
    is an integer.

    Args:
        dict_to_sort (dict): dict whose keys need sorting

    Returns:
        list: list of sorted keys from ``dict_to_sort``
    """
    sorted_keys = list(dict_to_sort.keys())
    sorted_keys.sort(key=lambda x: int(x[1:]))
    return sorted_keys



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for reading/writing files"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

import numpy as np
from skimage.io import imread, imsave
from tensorflow.keras import backend as K


def get_image(file_name):
    """DEPRECATED. Use ``skimage.io.imread`` instead.

    Read image from file and returns it as a tensor.

    Args:
        file_name (str): path to image file

    Returns:
        numpy.array: numpy array of image data
    """
    return np.float32(imread(file_name))


def save_model_output(output,
                      output_dir,
                      feature_name='',
                      channel=None,
                      data_format=None):
    """Save model output as tiff images in the provided directory

    Args:
        output (numpy.array): Output of a model.
            Expects channel to have its own axis.
        output_dir (str): Directory to save the model output images.
        feature_name (str): Optional description to start each output image
            filename.
        channel (int): If given, only saves this channel.
    """
    if data_format is None:
        data_format = K.image_data_format()
    channel_axis = 1 if data_format == 'channels_first' else -1
    z_axis = 2 if data_format == 'channels_first' else 1

    if channel is not None and not 0 < channel < output.shape[channel_axis]:
        raise ValueError('`channel` must be in the range of the output '
                         'channels. Got ', channel)

    if not os.path.isdir(output_dir):
        raise IOError('{} is not a valid output_dir'.format(
            output_dir))

    for b in range(output.shape[0]):
        # If multiple batches of results, create a numbered subdirectory
        batch_dir = str(b) if output.shape[0] > 1 else ''

        # If 2D, convert to 3D with only one z-axis
        if len(output.shape) == 4:
            output = np.expand_dims(output, axis=z_axis)

        for f in range(output.shape[z_axis]):
            for c in range(output.shape[channel_axis]):
                # if only saving one channel, skip the non-equal channels
                if channel is not None and channel != c:
                    continue

                if data_format == 'channels_first':
                    feature = output[b, c, f, :, :]
                else:
                    feature = output[b, f, :, :, c]

                zpad = max(3, len(str(output.shape[z_axis])))
                cnnout_name = 'feature_{}_frame_{}.tif'.format(c, str(f).zfill(zpad))
                if feature_name:
                    cnnout_name = '{}_{}'.format(feature_name, cnnout_name)

                out_file_path = os.path.join(output_dir, batch_dir, cnnout_name)
                imsave(out_file_path, feature.astype('int32'), check_contrast=False)
        print('Saved {} frames to {}'.format(output.shape[1], output_dir))



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for training neural nets"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
from tensorflow.keras import callbacks
from tensorflow.python.client import device_lib


def get_callbacks(model_path,
                  save_weights_only=False,
                  lr_sched=None,
                  tensorboard_log_dir=None,
                  reduce_lr_on_plateau=False,
                  monitor='val_loss',
                  verbose=1):
    """Returns a list of callbacks used for training

    Args:
        model_path: (str) path for the ``h5`` model file.
        save_weights_only: (bool) if True, then only the model's weights
            will be saved.
        lr_sched (function): learning rate scheduler per epoch.
            from `~deepcell.utils.train_utils.rate_scheduler`.
        tensorboard_log_dir (str): log directory for tensorboard.
        monitor (str): quantity to monitor.
        verbose (int): verbosity mode, 0 or 1.

    Returns:
        list: a list of callbacks to be passed to ``model.fit()``
    """
    cbs = [
        callbacks.ModelCheckpoint(
            model_path, monitor=monitor,
            save_best_only=True, verbose=verbose,
            save_weights_only=save_weights_only),
    ]

    if lr_sched:
        cbs.append(callbacks.LearningRateScheduler(lr_sched))

    if reduce_lr_on_plateau:
        cbs.append(
            callbacks.ReduceLROnPlateau(
                monitor=monitor, factor=0.1,
                patience=10, verbose=verbose,
                mode='auto', min_delta=0.0001,
                cooldown=0, min_lr=0)
        )

    if tensorboard_log_dir:
        cbs.append(callbacks.TensorBoard(log_dir=tensorboard_log_dir))

    return cbs


def rate_scheduler(lr=.001, decay=0.95):
    """Schedule the learning rate based on the epoch.

    Args:
        lr (float): initial learning rate
        decay (float): rate of decay of the learning rate

    Returns:
        function: A function that takes in the epoch
        and returns a learning rate.
    """
    def output_fn(epoch):
        epoch = np.int(epoch)
        new_lr = lr * (decay ** epoch)
        return new_lr
    return output_fn


def count_gpus():
    """Get the number of available GPUs.

    Returns:
        int: count of GPUs as integer
    """
    devices = device_lib.list_local_devices()
    gpus = [d for d in devices if d.name.lower().startswith('/device:gpu')]
    return len(gpus)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for padding layers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from absl.testing import parameterized

from tensorflow.python import keras
# from tensorflow.python.eager import context
from tensorflow.python.keras import keras_parameterized
from tensorflow.python.keras import testing_utils
from tensorflow.python.platform import test

from deepcell import layers


def _get_random_padding(dim):
    R = lambda: np.random.randint(low=0, high=9)
    return tuple([(R(), R()) for _ in range(dim)])


@keras_parameterized.run_all_keras_modes
class ReflectionPaddingTest(keras_parameterized.TestCase):

    def test_reflection_padding_2d(self):
        num_samples = 2
        stack_size = 2
        input_num_row = 4
        input_num_col = 5

        custom_objects = {'ReflectionPadding2D': layers.ReflectionPadding2D}
        ins1 = np.ones((num_samples, input_num_row, input_num_col, stack_size))
        ins2 = np.ones((num_samples, stack_size, input_num_row, input_num_col))
        data_formats = ['channels_first', 'channels_last']
        with tf.keras.utils.custom_object_scope(custom_objects):
            for data_format, inputs in zip(data_formats, [ins2, ins1]):
                # basic test
                testing_utils.layer_test(
                    layers.ReflectionPadding2D,
                    kwargs={'padding': (2, 2),
                            'data_format': data_format},
                    input_shape=inputs.shape)
                testing_utils.layer_test(
                    layers.ReflectionPadding2D,
                    kwargs={'padding': ((1, 2), (3, 4)),
                            'data_format': data_format},
                    input_shape=inputs.shape)

        # correctness test
        # with self.cached_session():
        #     layer = layers.ReflectionPadding2D(
        #         padding=(2, 2), data_format=data_format)
        #     layer.build(inputs.shape)
        #     output = layer(tf.keras.backend.variable(inputs))
        #     if context.executing_eagerly():
        #         np_output = output.numpy()
        #     else:
        #         np_output = tf.keras.backend.eval(output)
        #     if data_format == 'channels_last':
        #         for offset in [0, 1, -1, -2]:
        #             np.testing.assert_allclose(np_output[:, offset, :, :], 0.)
        #             np.testing.assert_allclose(np_output[:, :, offset, :], 0.)
        #         np.testing.assert_allclose(np_output[:, 2:-2, 2:-2, :], 1.)
        #     elif data_format == 'channels_first':
        #         for offset in [0, 1, -1, -2]:
        #             np.testing.assert_allclose(np_output[:, :, offset, :], 0.)
        #             np.testing.assert_allclose(np_output[:, :, :, offset], 0.)
        #         np.testing.assert_allclose(np_output[:, 2:-2, 2:-2, :], 1.)

        #     layer = layers.ReflectionPadding2D(
        #         padding=((1, 2), (3, 4)), data_format=data_format)
        #     layer.build(inputs.shape)
        #     output = layer(tf.keras.backend.variable(inputs))
        #     if context.executing_eagerly():
        #         np_output = output.numpy()
        #     else:
        #         np_output = tf.keras.backend.eval(output)
        #     if data_format == 'channels_last':
        #         for top_offset in [0]:
        #             np.testing.assert_allclose(np_output[:, top_offset, :, :], 0.)
        #         for bottom_offset in [-1, -2]:
        #             np.testing.assert_allclose(np_output[:, bottom_offset, :, :], 0.)
        #         for left_offset in [0, 1, 2]:
        #             np.testing.assert_allclose(np_output[:, :, left_offset, :], 0.)
        #         for right_offset in [-1, -2, -3, -4]:
        #             np.testing.assert_allclose(np_output[:, :, right_offset, :], 0.)
        #         np.testing.assert_allclose(np_output[:, 1:-2, 3:-4, :], 1.)
        #     elif data_format == 'channels_first':
        #         for top_offset in [0]:
        #             np.testing.assert_allclose(np_output[:, :, top_offset, :], 0.)
        #         for bottom_offset in [-1, -2]:
        #             np.testing.assert_allclose(np_output[:, :, bottom_offset, :], 0.)
        #         for left_offset in [0, 1, 2]:
        #             np.testing.assert_allclose(np_output[:, :, :, left_offset], 0.)
        #         for right_offset in [-1, -2, -3, -4]:
        #             np.testing.assert_allclose(np_output[:, :, :, right_offset], 0.)
        #         np.testing.assert_allclose(np_output[:, :, 1:-2, 3:-4], 1.)

        # test incorrect use
        with self.assertRaises(ValueError):
            layers.ReflectionPadding2D(padding=(1, 1, 1))
        with self.assertRaises(ValueError):
            layers.ReflectionPadding2D(padding=None)

    def test_reflection_padding_3d(self):
        num_samples = 2
        stack_size = 2
        input_len_dim1 = 4
        input_len_dim2 = 5
        input_len_dim3 = 3

        custom_objects = {'ReflectionPadding3D': layers.ReflectionPadding3D}
        inputs1 = np.ones((num_samples, input_len_dim1, input_len_dim2,
                           input_len_dim3, stack_size))
        inputs2 = np.ones((num_samples, stack_size, input_len_dim1,
                           input_len_dim2, input_len_dim3))
        data_formats = ['channels_first', 'channels_last']
        with tf.keras.utils.custom_object_scope(custom_objects):
            for data_format, inputs in zip(data_formats, [inputs2, inputs1]):
                # basic test
                testing_utils.layer_test(
                    layers.ReflectionPadding3D,
                    kwargs={'padding': (2, 2, 2),
                            'data_format': data_format},
                    input_shape=inputs.shape)

        # correctness test
        # with self.cached_session():
        #     layer = layers.ReflectionPadding3D(padding=(2, 2, 2))
        #     layer.build(inputs.shape)
        #     output = layer(tf.keras.backend.variable(inputs))
        #     if context.executing_eagerly():
        #         np_output = output.numpy()
        #     else:
        #         np_output = tf.keras.backend.eval(output)
        #     for offset in [0, 1, -1, -2]:
        #         np.testing.assert_allclose(np_output[:, offset, :, :, :], 0.)
        #         np.testing.assert_allclose(np_output[:, :, offset, :, :], 0.)
        #         np.testing.assert_allclose(np_output[:, :, :, offset, :], 0.)
        #     np.testing.assert_allclose(np_output[:, 2:-2, 2:-2, 2:-2, :], 1.)

        # test incorrect use
        with self.assertRaises(ValueError):
            layers.ReflectionPadding3D(padding=(1, 1))
        with self.assertRaises(ValueError):
            layers.ReflectionPadding3D(padding=None)


if __name__ == '__main__':
    test.main()



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for running functions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized

import numpy as np

from tensorflow import keras
from tensorflow.python.platform import test

from deepcell import layers
from deepcell import running


class RunningTests(test.TestCase, parameterized.TestCase):

    def test_get_cropped_input_shape(self):
        # test 2D images
        img_w, img_h = 30, 30
        batches = 5
        n_crops = 4
        field = 11
        # test channels_last
        keras.backend.set_image_data_format('channels_last')
        X = np.random.random((batches, img_w, img_h, 1))
        shape = running.get_cropped_input_shape(
            images=X,
            num_crops=n_crops,
            receptive_field=field)

        new_w = img_w // n_crops + (field - 1)
        new_h = img_h // n_crops + (field - 1)
        self.assertEqual(shape, (new_w, new_h, 1))

        # test channels_first
        X = np.random.random((batches, 1, img_w, img_h))
        shape = running.get_cropped_input_shape(
            images=X,
            num_crops=n_crops,
            receptive_field=field,
            data_format='channels_first')

        new_w = img_w // n_crops + (field - 1)
        new_h = img_h // n_crops + (field - 1)
        self.assertEqual(shape, (1, new_w, new_h))

        # test 3D images
        frames = 30

        # test channels_last
        X = np.random.random((batches, frames, img_w, img_h, 1))
        shape = running.get_cropped_input_shape(
            images=X,
            num_crops=n_crops,
            receptive_field=field,
            data_format='channels_last')

        new_w = img_w // n_crops + (field - 1)
        new_h = img_h // n_crops + (field - 1)
        self.assertEqual(shape, (frames, new_w, new_h, 1))

        # test channels_first
        X = np.random.random((batches, 1, frames, img_w, img_h))
        shape = running.get_cropped_input_shape(
            images=X,
            num_crops=n_crops,
            receptive_field=field,
            data_format='channels_first')

        new_w = img_w // n_crops + (field - 1)
        new_h = img_h // n_crops + (field - 1)
        self.assertEqual(shape, (1, frames, new_w, new_h))

    def test_get_padding_layers(self):
        # test reflection padding layer
        model = keras.models.Sequential()
        model.add(layers.ReflectionPadding2D(padding=(30, 30)))
        model.add(keras.layers.Dense(4))

        padded = running.get_padding_layers(model)
        self.assertEqual(len(padded), 1)

        # test zero padding layer
        model = keras.models.Sequential()
        model.add(keras.layers.ZeroPadding2D(padding=(30, 30)))
        model.add(keras.layers.Dense(4))

        padded = running.get_padding_layers(model)
        self.assertEqual(len(padded), 1)

        # test no padding layer
        model = keras.models.Sequential()
        model.add(keras.layers.Dense(4))

        padded = running.get_padding_layers(model)
        self.assertEqual(len(padded), 0)

    @parameterized.named_parameters([
        {
            'testcase_name': '2d_channels_last',
            'data_format': 'channels_last',
            'shape': (2, 32, 32, 1)
        }, {
            'testcase_name': '3d_channels_last',
            'data_format': 'channels_last',
            'shape': (2, 8, 32, 32, 1)
        }, {
            'testcase_name': '2d_channels_first',
            'data_format': 'channels_first',
            'shape': (2, 1, 32, 32)
        }, {
            'testcase_name': '3d_channels_first',
            'data_format': 'channels_first',
            'shape': (2, 1, 8, 32, 32)
        },
    ])
    def test_process_whole_image(self, data_format, shape):
        keras.backend.set_image_data_format(data_format)

        num_crops = 2
        receptive_field = 3
        features = 3

        images = np.ones(shape)

        input_shape = running.get_cropped_input_shape(
            images, num_crops,
            receptive_field=receptive_field,
            data_format=data_format)

        for padding in ['reflect', 'zero']:
            with self.cached_session():
                inputs = keras.layers.Input(shape=input_shape)
                outputs = layers.TensorProduct(features)(inputs)
                model = keras.models.Model(inputs=inputs,
                                           outputs=[outputs, outputs])

                output = running.process_whole_image(
                    model, images,
                    num_crops=num_crops,
                    receptive_field=receptive_field,
                    padding=padding)

                if data_format == 'channels_first':
                    expected_shape = tuple([images.shape[0], features] +
                                           list(images.shape[2:]))
                else:
                    expected_shape = tuple([images.shape[0]] +
                                           list(images.shape[1:-1]) +
                                           [features])

                self.assertEqual(output.shape, expected_shape)

        with self.assertRaises(ValueError):
            inputs = keras.layers.Input(shape=(3, 4, 5))
            outputs = layers.TensorProduct(features)(inputs)
            model = keras.models.Model(inputs=inputs, outputs=outputs)

            output = running.process_whole_image(
                model, images,
                num_crops=num_crops,
                receptive_field=receptive_field,
                padding='reflect')

        with self.assertRaises(ValueError):
            inputs = keras.layers.Input(shape=input_shape)
            outputs = layers.TensorProduct(features)(inputs)
            model = keras.models.Model(inputs=inputs, outputs=outputs)

            output = running.process_whole_image(
                model, images,
                num_crops=num_crops,
                receptive_field=receptive_field,
                padding=None)


if __name__ == '__main__':
    test.main()



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Deepcell Applications - Pre-trained models for specific functions"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from deepcell.applications.application import Application
from deepcell.applications.cytoplasm_segmentation import CytoplasmSegmentation
from deepcell.applications.mesmer import Mesmer
from deepcell.applications.multiplex_segmentation import MultiplexSegmentation
from deepcell.applications.nuclear_segmentation import NuclearSegmentation
from deepcell.applications.cell_tracking import CellTracking
from deepcell.applications.label_detection import LabelDetection
from deepcell.applications.scale_detection import ScaleDetection

del absolute_import
del division
del print_function



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Scale data generators.

These generators perform random zooming on the raw data, and yield each image
along with the zoom value of the image. This allows for models to learn the
scale factor of data, which can be used to help normalize the zoom factor of
new data.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

import numpy as np

from tensorflow.keras import backend as K
from tensorflow.keras.preprocessing.image import array_to_img
from tensorflow.keras.preprocessing.image import Iterator

from deepcell.image_generators import ImageFullyConvDataGenerator


class ScaleIterator(Iterator):
    """Iterator yielding data from Numpy arrays (``X`` and ``y``).

    Args:
        train_dict (dict): Consists of numpy arrays for ``X`` and ``y``.
        scale_generator (ScaleDataGenerator): For random transformations
            and normalization.
        batch_size (int): Size of a batch.
        shuffle (bool): Whether to shuffle the data between epochs.
        seed (int): Random seed for data shuffling.
        data_format (str): A string, one of ``channels_last`` (default)
            or ``channels_first``. The ordering of the dimensions in the
            inputs. ``channels_last`` corresponds to inputs with shape
            ``(batch, height, width, channels)`` while ``channels_first``
            corresponds to inputs with shape
            ``(batch, channels, height, width)``.
        save_to_dir (str): Optional directory where to save the pictures
            being yielded, in a viewable format. This is useful
            for visualizing the random transformations being
            applied, for debugging purposes.
        save_prefix (str): Prefix to use for saving sample
            images (if ``save_to_dir`` is set).
        save_format (str): Format to use for saving sample images
            (if ``save_to_dir`` is set).
    """

    def __init__(self,
                 train_dict,
                 scale_generator,
                 batch_size=1,
                 shuffle=False,
                 seed=None,
                 data_format='channels_last',
                 save_to_dir=None,
                 save_prefix='',
                 save_format='png'):
        X, y = train_dict['X'], train_dict['y']
        if X.shape[0] != y.shape[0]:
            raise ValueError('Training batches and labels should have the same'
                             'length. Found X.shape: {} y.shape: {}'.format(
                                 X.shape, y.shape))
        self.x = np.asarray(X, dtype=K.floatx())

        if self.x.ndim != 4:
            raise ValueError('Input data in `ImageFullyConvIterator` '
                             'should have rank 4. You passed an array '
                             'with shape', self.x.shape)

        self.y = np.ones((self.x.shape[0], 1), dtype=K.floatx())
        self.channel_axis = 3 if data_format == 'channels_last' else 1
        self.scale_generator = scale_generator
        self.data_format = data_format
        self.save_to_dir = save_to_dir
        self.save_prefix = save_prefix
        self.save_format = save_format
        super(ScaleIterator, self).__init__(
            self.x.shape[0], batch_size, shuffle, seed)

    def _get_batches_of_transformed_samples(self, index_array):
        batch_x = np.zeros(tuple([len(index_array)] + list(self.x.shape)[1:]),
                           dtype=self.x.dtype)
        batch_y = np.zeros(tuple([len(index_array)] + list(self.y.shape)[1:]),
                           dtype=self.y.dtype)

        for i, j in enumerate(index_array):
            x = self.x[j]
            y = self.y[j]

            x, y = self.scale_generator.random_transform(x.astype(K.floatx()), y)

            x = self.scale_generator.standardize(x)

            batch_x[i] = x
            batch_y[i] = y

        if self.save_to_dir:
            for i, j in enumerate(index_array):
                if self.data_format == 'channels_first':
                    img_x = np.expand_dims(batch_x[i, 0, ...], 0)
                else:
                    img_x = np.expand_dims(batch_x[i, ..., 0], -1)
                img = array_to_img(img_x, self.data_format, scale=True)
                fname = '{prefix}_{index}_{hash}.{format}'.format(
                    prefix=self.save_prefix,
                    index=j,
                    hash=np.random.randint(1e4),
                    format=self.save_format)
                img.save(os.path.join(self.save_to_dir, fname))

        return batch_x, batch_y

    def next(self):
        """For python 2.x. Returns the next batch.
        """
        # Keeps under lock only the mechanism which advances
        # the indexing of each batch.
        with self.lock:
            index_array = next(self.index_generator)
        # The transformation of images is not under thread lock
        # so it can be done in parallel
        return self._get_batches_of_transformed_samples(index_array)


class ScaleDataGenerator(ImageFullyConvDataGenerator):
    """Generates batches of tensor image data with real-time data augmentation.
    The data will be looped over (in batches).

    Args:
        featurewise_center (bool): Set input mean to 0 over the dataset,
            feature-wise.
        samplewise_center (bool): Set each sample mean to 0.
        featurewise_std_normalization (bool): Divide inputs by std
            of the dataset, feature-wise.
        samplewise_std_normalization (bool): Divide each input by its std.
        zca_epsilon (float): Epsilon for ZCA whitening. Default is 1e-6.
        zca_whitening (bool): Apply ZCA whitening.
        rotation_range (int): Degree range for random rotations.
        width_shift_range (float): 1-D array-like or int

            - float: fraction of total width, if < 1, or pixels if >= 1.
            - 1-D array-like: random elements from the array.
            - int: integer number of pixels from interval
              ``(-width_shift_range, +width_shift_range)``
            - With ``width_shift_range=2`` possible values are integers
              ``[-1, 0, +1]``, same as with ``width_shift_range=[-1, 0, +1]``,
              while with ``width_shift_range=1.0`` possible values are floats
              in the interval [-1.0, +1.0).

        height_shift_range: Float, 1-D array-like or int

            - float: fraction of total height, if < 1, or pixels if >= 1.
            - 1-D array-like: random elements from the array.
            - int: integer number of pixels from interval
              ``(-height_shift_range, +height_shift_range)``
            - With ``height_shift_range=2`` possible values
              are integers ``[-1, 0, +1]``,
              same as with ``height_shift_range=[-1, 0, +1]``,
              while with ``height_shift_range=1.0`` possible values are floats
              in the interval [-1.0, +1.0).

        shear_range (float): Shear Intensity
            (Shear angle in counter-clockwise direction in degrees)
        zoom_range (float): float or [lower, upper], Range for random zoom.
            If a float, ``[lower, upper] = [1-zoom_range, 1+zoom_range]``.
        channel_shift_range (float): range for random channel shifts.
        fill_mode (str): One of {"constant", "nearest", "reflect" or "wrap"}.

            Default is 'nearest'. Points outside the boundaries of the input
            are filled according to the given mode:

                - 'constant': kkkkkkkk|abcd|kkkkkkkk (cval=k)
                - 'nearest':  aaaaaaaa|abcd|dddddddd
                - 'reflect':  abcddcba|abcd|dcbaabcd
                - 'wrap':  abcdabcd|abcd|abcdabcd

        cval (float): Value used for points outside the boundaries
            when ``fill_mode = "constant"``.
        horizontal_flip (bool): Randomly flip inputs horizontally.
        vertical_flip (bool): Randomly flip inputs vertically.
        rescale: rescaling factor. Defaults to None. If None or 0, no rescaling
            is applied, otherwise we multiply the data by the value provided
            (before applying any other transformation).
        preprocessing_function: function that will be implied on each input.
            The function will run after the image is resized and augmented.
            The function should take one argument:
            one image (Numpy tensor with rank 3),
            and should output a Numpy tensor with the same shape.
        data_format (str): A string, one of ``channels_last`` (default)
            or ``channels_first``. The ordering of the dimensions in the
            inputs. ``channels_last`` corresponds to inputs with shape
            ``(batch, height, width, channels)`` while ``channels_first``
            corresponds to inputs with shape
            ``(batch, channels, height, width)``.
        validation_split (float): Fraction of images reserved for validation
            (strictly between 0 and 1).
    """

    def flow(self,
             train_dict,
             batch_size=1,
             shuffle=True,
             seed=None,
             save_to_dir=None,
             save_prefix='',
             save_format='png'):
        """Generates batches of augmented/normalized data with given arrays.

        Args:
            train_dict (dict): Consists of numpy arrays for ``X`` and ``y``.
            batch_size (int): Size of a batch.
            shuffle (bool): Whether to shuffle the data between epochs.
            seed (int): Random seed for data shuffling.
            save_to_dir (str): Optional directory where to save the pictures
                being yielded, in a viewable format. This is useful
                for visualizing the random transformations being
                applied, for debugging purposes.
            save_prefix (str): Prefix to use for saving sample
                images (if ``save_to_dir`` is set).
            save_format (str): Format to use for saving sample images
                (if ``save_to_dir`` is set).

        Returns:
            ScaleIterator: An ``Iterator`` yielding tuples of ``(x, y)``,
            where ``x`` is a numpy array of image data and ``y``
            is the transformed scale of ``x``.
        """
        return ScaleIterator(
            train_dict,
            self,
            batch_size=batch_size,
            shuffle=shuffle,
            seed=seed,
            data_format=self.data_format,
            save_to_dir=save_to_dir,
            save_prefix=save_prefix,
            save_format=save_format)

    def random_transform(self, x, y=None, seed=None):
        """Applies a random transformation to an image.

        Args:
            x (tensor): 3D tensor or list of 3D tensors,
                single image.
            y (tensor): 3D tensor or list of 3D tensors,
                label mask(s) for x, optional.
            seed (int): Random seed.

        Returns:
            tensor:  A randomly transformed version of the input (same shape).
            If ``y`` is passed, it is transformed if necessary and returned.
        """
        params = self.get_random_transform(x.shape, seed)
        params['zy'] = params['zx']

        if isinstance(x, list):
            x = [self.apply_transform(x_i, params) for x_i in x]
        else:
            x = self.apply_transform(x, params)

        if y is None:
            return x
        return x, np.array(params['zx'])



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Layers to generate tensor products for 2D and 3D data"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import tensorflow as tf
from tensorflow.python.framework import tensor_shape
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import InputSpec
from tensorflow.python.keras.utils import conv_utils


class TensorProduct(Layer):
    """Just your regular densely-connected NN layer.

    Dense implements the operation:

    ``output = activation(dot(input, kernel) + bias)``

    where ``activation`` is the element-wise activation function
    passed as the ``activation`` argument, ``kernel`` is a weights matrix
    created by the layer, and ``bias`` is a bias vector created by the layer
    (only applicable if ``use_bias`` is ``True``).

    Note: if the input to the layer has a rank greater than 2, then
    it is flattened prior to the initial dot product with ``kernel``.

    Args:
        output_dim (int): Positive integer, dimensionality of the output space.
        data_format (str): A string, one of ``channels_last`` (default)
            or ``channels_first``. The ordering of the dimensions in the
            inputs. ``channels_last`` corresponds to inputs with shape
            ``(batch, height, width, channels)`` while ``channels_first``
            corresponds to inputs with shape
            ``(batch, channels, height, width)``.
        activation (function): Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: ``a(x) = x``).
        use_bias (bool): Whether the layer uses a bias.
        kernel_initializer (function): Initializer for the ``kernel`` weights
            matrix, used for the linear transformation of the inputs.
        bias_initializer (function): Initializer for the bias vector. If None,
            the default initializer will be used.
        kernel_regularizer (function): Regularizer function applied to the
            ``kernel`` weights matrix.
        bias_regularizer (function): Regularizer function applied to the
            bias vector.
        activity_regularizer (function): Regularizer function applied to.
        kernel_constraint (function): Constraint function applied to
            the ``kernel`` weights matrix.
        bias_constraint (function): Constraint function applied to the
            bias vector.

    Input shape:
        nD tensor with shape: (batch_size, ..., input_dim).
        The most common situation would be
        a 2D input with shape (batch_size, input_dim).

    Output shape:
        nD tensor with shape: (batch_size, ..., output_dim).
        For instance, for a 2D input with shape (batch_size, input_dim),
        the output would have shape (batch_size, output_dim).
    """

    def __init__(self,
                 output_dim,
                 data_format=None,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)

        super(TensorProduct, self).__init__(
            activity_regularizer=tf.keras.regularizers.get(
                activity_regularizer), **kwargs)

        self.output_dim = int(output_dim)
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.activation = tf.keras.activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.activity_regularizer = tf.keras.regularizers.get(activity_regularizer)
        self.kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)

        self.supports_masking = True
        self.input_spec = InputSpec(min_ndim=2)

    def build(self, input_shape):
        if self.data_format == 'channels_first':
            channel_axis = 1
        else:
            channel_axis = -1
        input_shape = tensor_shape.TensorShape(input_shape)
        if input_shape.dims[channel_axis].value is None:
            raise ValueError('The channel dimension of the inputs to '
                             '`TensorProduct` should be defined. '
                             'Found `None`.')
        input_dim = int(input_shape[channel_axis])
        kernel_shape = (input_dim, self.output_dim)

        self.kernel = self.add_weight(
            'kernel',
            shape=kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            dtype=self.dtype,
            trainable=True)
        if self.use_bias:
            self.bias = self.add_weight(
                'bias',
                shape=(self.output_dim,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                dtype=self.dtype,
                trainable=True)
        else:
            self.bias = None

        # Set input spec.
        self.input_spec = InputSpec(min_ndim=2,
                                    axes={channel_axis: input_dim})
        self.built = True

    def call(self, inputs):
        rank = len(inputs.get_shape().as_list())
        if self.data_format == 'channels_first':
            pattern = [0, rank - 1] + list(range(1, rank - 1))
            output = tf.tensordot(inputs, self.kernel, axes=[[1], [0]])
            output = K.permute_dimensions(output, pattern=pattern)
            # output = K.dot(inputs, self.kernel)

        elif self.data_format == 'channels_last':
            output = tf.tensordot(inputs, self.kernel, axes=[[rank - 1], [0]])

        if self.use_bias:
            output = K.bias_add(output, self.bias, self.data_format)

        if self.activation is not None:
            return self.activation(output)

        return output

    def compute_output_shape(self, input_shape):
        input_shape = tensor_shape.TensorShape(input_shape).as_list()
        if self.data_format == 'channels_first':
            output_shape = tuple([input_shape[0], self.output_dim] +
                                 list(input_shape[2:]))
        else:
            output_shape = tuple(list(input_shape[:-1]) + [self.output_dim])

        return tensor_shape.TensorShape(output_shape)

    def get_config(self):
        config = {
            'output_dim': self.output_dim,
            'data_format': self.data_format,
            'activation': self.activation,
            'use_bias': self.use_bias,
            'kernel_initializer': tf.keras.initializers.serialize(
                self.kernel_initializer),
            'bias_initializer': tf.keras.initializers.serialize(
                self.bias_initializer),
            'kernel_regularizer': tf.keras.regularizers.serialize(
                self.kernel_regularizer),
            'bias_regularizer': tf.keras.regularizers.serialize(
                self.bias_regularizer),
            'activity_regularizer': tf.keras.regularizers.serialize(
                self.activity_regularizer),
            'kernel_constraint': tf.keras.constraints.serialize(
                self.kernel_constraint),
            'bias_constraint': tf.keras.constraints.serialize(
                self.bias_constraint)
        }
        base_config = super(TensorProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
